#include "bsde_internal.h"

static BSDE_Contact *bsde_contactpool=NULL;
static int bsde_contactpool_pos=0;

BSDE_Contact *BSDE_MakeContact(
	BSDE_World *world,
	BSDE_Solid *obja, BSDE_Solid *objb,
	double *org, double *norm, double depth)
{
	BSDE_Contact *tmp;

	if(!bsde_contactpool)
	{
		bsde_contactpool=malloc(BSDE_MAX_CONTACTS*
			sizeof(BSDE_Contact));
		bsde_contactpool_pos=0;
	}

	tmp=&bsde_contactpool[bsde_contactpool_pos++];

	tmp->obja=obja;
	tmp->objb=objb;
	tmp->depth=depth;
	V3D_COPY(org, tmp->org);
	V3D_COPY(norm, tmp->norm);

	return(tmp);
}

void BSDE_FreeContacts()
{
	bsde_contactpool_pos=0;
}

BSDE_Solid *BSDE_NewSolidArray(int type, int **arr,
	double **varr, double **sarr)
{
	BSDE_Solid *tmp;

	switch(type)
	{
	case BSDE_SOLID_BSP:
		tmp=BSDE_SolidBSP_BuildArray(arr, varr);
		break;
	default:
		tmp=NULL;
		break;
	}
	return(tmp);
}

void BSDE_ArraySkipUnknown(int **arr, double **varr)
{
	switch(**arr)
	{
	case BSDE_TAG_VERTEX:
		*arr=*arr+1;
		*varr=*varr+4;
		break;
	case BSDE_TAG_NORMAL:
		*arr=*arr+1;
		*varr=*varr+4;
		break;
	case BSDE_TAG_BEGIN:
		*arr=*arr+2;
		*varr=*varr+4;
		while(**arr!=BSDE_TAG_END)
			BSDE_ArraySkipUnknown(arr, varr);
		*arr=*arr+1;
		break;
	defualt:
		printf("BSDE_ArraySkipUnknown: Unknown tag %d\n", **arr);
		*arr=*arr+1;
		break;
	}
}

BSDE_Solid *BSDE_NewSolid(int type)
{
	BSDE_Solid *tmp;

	tmp=malloc(sizeof(BSDE_Solid));
	memset(tmp, 0, sizeof(BSDE_Solid));
	tmp->solidtype=type;
	BSDE_QuatD_Identity(tmp->rot);

	V3D_CONST(tmp->mins, -1);
	V3D_CONST(tmp->maxs, 1);
	tmp->radius=1;
	tmp->height=1;
	tmp->volume=8;
	tmp->mass=8;

	tmp->elastic=BSDE_M_DEF_ELASTIC;
	tmp->friction=BSDE_M_DEF_FRICTION;

	return(tmp);
}

BSDE_Solid *BSDE_NewAABB(
	double mx, double my, double mz,
	double nx, double ny, double nz)
{
	BSDE_Solid *tmp;

	tmp=malloc(sizeof(BSDE_Solid));
	memset(tmp, 0, sizeof(BSDE_Solid));
	tmp->solidtype=BSDE_SOLID_AABB;
	BSDE_QuatD_Identity(tmp->rot);

	V3F_SET(tmp->mins, mx, my, mz);
	V3F_SET(tmp->maxs, nx, ny, nz);
	tmp->volume=(nx-mx)*(ny-my)*(nz-mz);
	tmp->mass=tmp->volume;

	tmp->elastic=BSDE_M_DEF_ELASTIC;
	tmp->friction=BSDE_M_DEF_FRICTION;
	BSDE_BuildInertiaTensor(tmp);

	return(tmp);
}

BSDE_Solid *BSDE_NewOBB(
	double mx, double my, double mz,
	double nx, double ny, double nz)
{
	BSDE_Solid *tmp;

	tmp=BSDE_NewAABB(mx, my, mz, nx, ny, nz);
	tmp->solidtype=BSDE_SOLID_OBB;
	return(tmp);
}

BSDE_Solid *BSDE_NewSphere(double rad)
{
	BSDE_Solid *tmp;

	tmp=malloc(sizeof(BSDE_Solid));
	memset(tmp, 0, sizeof(BSDE_Solid));
	tmp->solidtype=BSDE_SOLID_SPHERE;
	BSDE_QuatD_Identity(tmp->rot);

	tmp->radius=rad;
	tmp->volume=(4.0/3.0)*BSDE_M_PI*(rad*rad);
	tmp->mass=tmp->volume;

	tmp->elastic=BSDE_M_DEF_ELASTIC;
	tmp->friction=BSDE_M_DEF_FRICTION;
	BSDE_BuildInertiaTensor(tmp);

	return(tmp);
}

BSDE_Solid *BSDE_NewCylinder(double height, double radius)
{
	BSDE_Solid *tmp;

	tmp=malloc(sizeof(BSDE_Solid));
	memset(tmp, 0, sizeof(BSDE_Solid));
	tmp->solidtype=BSDE_SOLID_CYLINDER;
	BSDE_QuatD_Identity(tmp->rot);

	tmp->height=height;
	tmp->radius=radius;
	tmp->volume=BSDE_M_PI*(radius*radius)*height;
	tmp->mass=tmp->volume;

	tmp->elastic=BSDE_M_DEF_ELASTIC;
	tmp->friction=BSDE_M_DEF_FRICTION;
	BSDE_BuildInertiaTensor(tmp);

	return(tmp);
}

BSDE_Solid *BSDE_NewCappedCylinder(double height, double radius)
{
	BSDE_Solid *tmp;

	tmp=malloc(sizeof(BSDE_Solid));
	memset(tmp, 0, sizeof(BSDE_Solid));
	tmp->solidtype=BSDE_SOLID_CCYLINDER;
	BSDE_QuatD_Identity(tmp->rot);

	tmp->height=height;
	tmp->radius=radius;

	tmp->volume=((4.0/3.0)+height)*BSDE_M_PI*(radius*radius);
	tmp->mass=tmp->volume;
	BSDE_BuildInertiaTensor(tmp);

	tmp->elastic=BSDE_M_DEF_ELASTIC;
	tmp->friction=BSDE_M_DEF_FRICTION;

	return(tmp);
}

BSDE_Solid *BSDE_NewPlane(double x, double y, double z, double d)
{
	BSDE_Solid *tmp;

	tmp=malloc(sizeof(BSDE_Solid));
	memset(tmp, 0, sizeof(BSDE_Solid));
	tmp->solidtype=BSDE_SOLID_PLANE;
	BSDE_QuatD_Identity(tmp->rot);

	tmp->dir[0]=x;
	tmp->dir[1]=y;
	tmp->dir[2]=z;
	tmp->dir[3]=d;

	V3D_NORMALIZE(tmp->dir, tmp->dir);

	//so there is something sane here...
	tmp->volume=999999999999.0;
	tmp->mass=999999999999.0;

	tmp->elastic=BSDE_M_DEF_ELASTIC;
	tmp->friction=BSDE_M_DEF_FRICTION;

	return(tmp);
}

BSDE_Solid *BSDE_NewHull(double *norms, int n)
{
	BSDE_Solid *tmp;
	int i;

	tmp=malloc(sizeof(BSDE_Solid));
	memset(tmp, 0, sizeof(BSDE_Solid));
	tmp->solidtype=BSDE_SOLID_HULL;
	BSDE_QuatD_Identity(tmp->rot);

	tmp->norms=malloc(4*n*sizeof(double));
	tmp->n_faces=n;
	for(i=0; i<n; i++)
	{
		V3D_NORMALIZE(norms+i*4, tmp->norms+i*4);
		tmp->norms[i*4+3]=norms[i*4+3];
	}

	BSDE_SolidHull_MakeHullFaces(tmp);
	tmp->mass=tmp->volume;

	BSDE_BuildInertiaTensor(tmp);

#if 0
	tmp->mins[0]=-1;
	tmp->mins[1]=-1;
	tmp->mins[2]=-1;
	tmp->maxs[0]=1;
	tmp->maxs[1]=1;
	tmp->maxs[2]=1;
	BSDE_BuildBoxInertiaTensor(tmp);
#endif

	tmp->elastic=BSDE_M_DEF_ELASTIC;
	tmp->friction=BSDE_M_DEF_FRICTION;

	return(tmp);
}

BSDE_Solid *BSDE_NewTriMesh(double *vecs, int *tris, int nv, int nt)
{
	BSDE_Solid *tmp;
	double tv[3], tv0[3], tv1[3], tv2[3];
	double *v0, *v1, *v2;
	int i;

	tmp=malloc(sizeof(BSDE_Solid));
	memset(tmp, 0, sizeof(BSDE_Solid));
	tmp->solidtype=BSDE_SOLID_TRIMESH;
	BSDE_QuatD_Identity(tmp->rot);

	tmp->vecs=malloc(3*nv*sizeof(double));
	tmp->faces=malloc(3*nt*sizeof(int));
	tmp->norms=malloc(4*nt*sizeof(double));
	tmp->evecs=malloc(3*4*nt*sizeof(double));
	tmp->n_vecs=nv;
	tmp->n_faces=nt;

	V3D_INITBOX(tmp->mins, tmp->maxs);
	for(i=0; i<nv; i++)
	{
		V3D_COPY(vecs+i*3, tmp->vecs+i*3);

		V3D_COPY(vecs+i*3, tv);
		V3D_MIN(tmp->mins, tv, tmp->mins);
		V3D_MAX(tmp->maxs, tv, tmp->maxs);
	}

	for(i=0; i<nt; i++)
	{
		tmp->faces[i*3+0]=tris[i*3+0];
		tmp->faces[i*3+1]=tris[i*3+1];
		tmp->faces[i*3+2]=tris[i*3+2];

		v0=vecs+tris[i*3+0]*3;
		v1=vecs+tris[i*3+1]*3;
		v2=vecs+tris[i*3+2]*3;

		V3D_SUB(v1, v0, tv0);
		V3D_SUB(v2, v0, tv1);
		V3D_CROSS(tv2, tv1, tv);

		V3D_NORMALIZE(tv, tmp->norms+i*4);
		tmp->norms[i*4+3]=V3D_DOT(v0, tmp->norms+i*4);

		V3D_SUB(v1, v0, tv0);
		V3D_SUB(v2, v1, tv1);
		V3D_SUB(v0, v2, tv2);

		V3D_CROSS(tv0, tmp->evecs+i*12+0*4, tv);
		tmp->norms[i*12+0*4+3]=V3D_DOT(v0, tmp->evecs+i*12+0*4);

		V3D_CROSS(tv1, tmp->evecs+i*12+1*4, tv);
		tmp->norms[i*12+1*4+3]=V3D_DOT(v1, tmp->evecs+i*12+1*4);

		V3D_CROSS(tv2, tmp->evecs+i*12+2*4, tv);
		tmp->norms[i*12+2*4+3]=V3D_DOT(v2, tmp->evecs+i*12+2*4);
	}

	tmp->volume=BSDE_SolidBRep_ApproxBrepVolume(vecs, tris, nv, nt);
	tmp->mass=tmp->volume;

	BSDE_BuildInertiaTensor(tmp);

	tmp->elastic=BSDE_M_DEF_ELASTIC;
	tmp->friction=BSDE_M_DEF_FRICTION;

	return(tmp);
}


void BSDE_Destroy(BSDE_Solid *self)
{
	//FIXME: better handle bsp's, hulls, ...
	if(self->vecs)free(self->vecs);
	if(self->norms)free(self->norms);
	if(self->faces)free(self->faces);
	if(self->facevecs)free(self->facevecs);

	free(self);
}

void BSDE_Box(BSDE_Solid *self, double *mins, double *maxs)
{
	switch(self->solidtype)
	{
	case BSDE_SOLID_AABB:
		V3D_ADD(self->org, self->mins, mins);
		V3D_ADD(self->org, self->maxs, maxs);
		break;
	case BSDE_SOLID_OBB:
		BSDE_SolidOBB_Box(self->mins, self->maxs,
			self->xform, mins, maxs);
		break;
	case BSDE_SOLID_SPHERE:
		V3D_SUBCONST(self->org, self->radius, mins);
		V3D_ADDCONST(self->org, self->radius, maxs);
		break;
	case BSDE_SOLID_CYLINDER:
		BSDE_SolidCylinder_MakeBox(self->height, self->radius, 0,
			self->xform, mins, maxs);
		break;
	case BSDE_SOLID_CCYLINDER:
		BSDE_SolidCylinder_MakeBox(self->height, self->radius, 1,
			self->xform, mins, maxs);
		break;

	case BSDE_SOLID_PLANE:
		V3D_CONST(mins, -999999999);
		V3D_CONST(maxs, 999999999);
		break;

	case BSDE_SOLID_TRIMESH:	//inaccurate
		BSDE_SolidOBB_Box(self->mins, self->maxs,
			self->xform, mins, maxs);
		break;
	case BSDE_SOLID_HULL:
		BSDE_SolidHull_Box(self, self->xform, mins, maxs);
		break;
	case BSDE_SOLID_BSP:
		V3D_ADD(self->org, self->mins, mins);
		V3D_ADD(self->org, self->maxs, maxs);
		break;

	default:
		break;
	}
}

void BSDE_BoxMove(BSDE_Solid *self, double *mins, double *maxs)
{
	double f, g;
	int i;

	switch(self->solidtype)
	{
	case BSDE_SOLID_AABB:
		for(i=0; i<3; i++)
		{
			f=self->mins[i]+self->org[i];
			g=self->mins[i]+self->e_org[i];
			mins[i]=(f<g)?f:g;

			f=self->maxs[i]+self->org[i];
			g=self->maxs[i]+self->e_org[i];
			maxs[i]=(f>g)?f:g;
		}
		break;
	case BSDE_SOLID_OBB:
		BSDE_SolidOBB_BoxMove(self->mins, self->maxs,
			self->xform, self->e_xform, mins, maxs);
		break;
	case BSDE_SOLID_SPHERE:
		for(i=0; i<3; i++)
		{
			f=self->org[i]-self->radius;
			g=self->e_org[i]-self->radius;
			mins[i]=(f<g)?f:g;

			f=self->org[i]+self->radius;
			g=self->e_org[i]+self->radius;
			maxs[i]=(f>g)?f:g;
		}
		break;
	case BSDE_SOLID_CYLINDER:
		BSDE_SolidCylinder_BoxMove(self->height, self->radius, 0,
			self->xform, self->e_xform, mins, maxs);
		break;
	case BSDE_SOLID_CCYLINDER:
		BSDE_SolidCylinder_BoxMove(self->height, self->radius, 1,
			self->xform, self->e_xform, mins, maxs);
		break;
	case BSDE_SOLID_PLANE:
		V3D_CONST(mins, -999999999);
		V3D_CONST(maxs, 999999999);
		break;

	case BSDE_SOLID_TRIMESH:
		BSDE_SolidOBB_BoxMove(self->mins, self->maxs,
			self->xform, self->e_xform, mins, maxs);
		break;
	case BSDE_SOLID_HULL:
		BSDE_SolidHull_BoxMove(self, self->xform, self->e_xform,
			mins, maxs);
		break;
	case BSDE_SOLID_BSP:
		V3D_ADD(self->org, self->mins, mins);
		V3D_ADD(self->org, self->maxs, maxs);
		break;
	default:
		break;
	}
}

void BSDE_BoxMovePos(BSDE_Solid *self,
	double *sxform, double *exform,
	double *sorg, double *eorg,
	double *mins, double *maxs)
{
	double f, g;
	int i;

	switch(self->solidtype)
	{
	case BSDE_SOLID_AABB:
		for(i=0; i<3; i++)
		{
			f=self->mins[i]+sorg[i];
			g=self->mins[i]+eorg[i];
			mins[i]=(f<g)?f:g;

			f=self->maxs[i]+sorg[i];
			g=self->maxs[i]+eorg[i];
			maxs[i]=(f>g)?f:g;
		}
		break;
	case BSDE_SOLID_OBB:
		BSDE_SolidOBB_BoxMove(self->mins, self->maxs,
			sxform, exform, mins, maxs);
		break;
	case BSDE_SOLID_SPHERE:
		for(i=0; i<3; i++)
		{
			f=sorg[i]-self->radius;
			g=eorg[i]-self->radius;
			mins[i]=(f<g)?f:g;

			f=sorg[i]+self->radius;
			g=eorg[i]+self->radius;
			maxs[i]=(f>g)?f:g;
		}
		break;
	case BSDE_SOLID_CYLINDER:
		BSDE_SolidCylinder_BoxMove(self->height, self->radius, 0,
			sxform, exform, mins, maxs);
		break;
	case BSDE_SOLID_CCYLINDER:
		BSDE_SolidCylinder_BoxMove(self->height, self->radius, 1,
			sxform, exform, mins, maxs);
		break;
	case BSDE_SOLID_PLANE:
		V3D_CONST(mins, -999999999);
		V3D_CONST(maxs, 999999999);
		break;

	case BSDE_SOLID_TRIMESH:
		BSDE_SolidOBB_BoxMove(self->mins, self->maxs,
			sxform, exform, mins, maxs);
		break;
	case BSDE_SOLID_HULL:
		BSDE_SolidHull_BoxMove(self, sxform, exform, mins, maxs);
		break;
	case BSDE_SOLID_BSP:
		for(i=0; i<3; i++)
		{
			f=self->mins[i]+sorg[i];
			g=self->mins[i]+eorg[i];
			mins[i]=(f<g)?f:g;

			f=self->maxs[i]+sorg[i];
			g=self->maxs[i]+eorg[i];
			maxs[i]=(f>g)?f:g;
		}
		break;
	default:
		break;
	}
}


double BSDE_GetAttrF(BSDE_Solid *self, int var)
{
	double f;

	f=0;
	switch(var)
	{
	case BSDE_HEIGHT:
		f=self->height;
		break;
	case BSDE_RADIUS:
		f=self->radius;
		break;

	case BSDE_MASS:
		f=self->mass;
		break;
	case BSDE_VOLUME:
		f=self->volume;
		break;
	case BSDE_DENSITY:
		f=self->mass/self->volume;
		break;
	case BSDE_ELASTIC:
		f=self->elastic;
		break;
	case BSDE_FRICTION:
		f=self->friction;
		break;
	case BSDE_VISCOSITY:
		f=self->friction;
		break;
	case BSDE_CONTACT_POWER:
		f=self->u_wpow;
		break;
	default:
		break;
	}
	return(f);
}

void BSDE_SetAttrF(BSDE_Solid *self, int var, double val)
{
	switch(var)
	{
	case BSDE_HEIGHT:
		self->height=val;
		break;
	case BSDE_RADIUS:
		self->radius=val;
		break;

	case BSDE_MASS:
		self->mass=val;
		break;
	case BSDE_VOLUME:
		self->volume=val;
		break;
	case BSDE_DENSITY:
		self->mass=val*self->volume;
		break;
	case BSDE_ELASTIC:
		self->elastic=val;
		break;
	case BSDE_FRICTION:
		self->friction=val;
		break;
	case BSDE_VISCOSITY:
		self->viscosity=val;
		break;
	case BSDE_CONTACT_POWER:
		self->u_wpow=val;
		break;
	default:
		break;
	}
}

int BSDE_GetAttrI(BSDE_Solid *self, int var)
{
	int i;

	i=0;
	switch(var)
	{
	case BSDE_SOLIDTYPE:
		i=self->solidtype;
		break;
	case BSDE_NOSPIN:
		i=(self->moveflags&BSDE_MOVEFL_NOSPIN)!=0;
		break;
	case BSDE_NOMOVE:
		i=(self->moveflags&BSDE_MOVEFL_NOMOVE)!=0;
		break;
	case BSDE_NOGRAVITY:
		i=(self->moveflags&BSDE_MOVEFL_NOGRAVITY)!=0;
		break;
	case BSDE_STATIC:
		i=(self->moveflags&BSDE_MOVEFL_STATIC)!=0;
		break;
	case BSDE_SEMISTATIC:
		i=(self->moveflags&BSDE_MOVEFL_SEMISTATIC)!=0;
		break;
	case BSDE_FLUID:
		i=(self->moveflags&BSDE_MOVEFL_FLUID)!=0;
		break;
	case BSDE_NONSOLID:
		i=(self->moveflags&BSDE_MOVEFL_NONSOLID)!=0;
		break;
	case BSDE_TRIGGER:
		i=(self->moveflags&BSDE_MOVEFL_TRIGGER)!=0;
		break;

	case BSDE_NUM_FACES:
		i=self->n_faces;
		break;
	case BSDE_NUM_VERTICES:
		i=self->n_vecs;
		break;

	default:
		break;
	}
	return(i);
}

void BSDE_SetAttrI(BSDE_Solid *self, int var, int val)
{
	int i;
	switch(var)
	{
	case BSDE_NOSPIN:
		if(val)self->moveflags|=BSDE_MOVEFL_NOSPIN;
			else self->moveflags&=~BSDE_MOVEFL_NOSPIN;
		break;
	case BSDE_NOMOVE:
		if(val)self->moveflags|=BSDE_MOVEFL_NOMOVE;
			else self->moveflags&=~BSDE_MOVEFL_NOMOVE;
		break;
	case BSDE_NOGRAVITY:
		if(val)self->moveflags|=BSDE_MOVEFL_NOGRAVITY;
			else self->moveflags&=~BSDE_MOVEFL_NOGRAVITY;
		break;
	case BSDE_STATIC:
		i=BSDE_MOVEFL_NOSPIN|BSDE_MOVEFL_NOMOVE|
			BSDE_MOVEFL_NOGRAVITY|BSDE_MOVEFL_STATIC;
		if(val)self->moveflags|=i;
			else self->moveflags&=~i;
		break;
	case BSDE_SEMISTATIC:
		i=BSDE_MOVEFL_NOSPIN|BSDE_MOVEFL_NOMOVE|
			BSDE_MOVEFL_NOGRAVITY|BSDE_MOVEFL_SEMISTATIC;
		if(val)self->moveflags|=i;
			else self->moveflags&=~i;
		break;

	case BSDE_FLUID:
		if(val)self->moveflags|=BSDE_MOVEFL_FLUID;
			else self->moveflags&=~BSDE_MOVEFL_FLUID;
		break;
	case BSDE_NONSOLID:
		if(val)self->moveflags|=BSDE_MOVEFL_NONSOLID;
			else self->moveflags&=~BSDE_MOVEFL_NONSOLID;
		break;
	case BSDE_TRIGGER:
		if(val)self->moveflags|=BSDE_MOVEFL_TRIGGER;
			else self->moveflags&=~BSDE_MOVEFL_TRIGGER;
		break;
	default:
		break;
	}
}

void BSDE_GetAttrFV(BSDE_Solid *self, int var, float *val, int n)
{
	int i, j;

	switch(var)
	{
	case BSDE_ORIGIN:
		V3D_COPY(self->org, val);
		break;
	case BSDE_VELOCITY:
		V3D_COPY(self->vel, val);
		break;
	case BSDE_TORQUE:
		V3D_COPY(self->tvel, val);
		break;
	case BSDE_ANGULAR_VELOCITY:
		V3D_COPY(self->avel, val);
		break;
	case BSDE_ROTATION:
		V4D_COPY(self->rot, val);
		break;

	case BSDE_MINS:
		V3D_COPY(self->mins, val);
		break;
	case BSDE_MAXS:
		V3D_COPY(self->maxs, val);
		break;

	case BSDE_MASS_OFFSET:
		V3D_COPY(self->mass_offs, val);
		break;
	case BSDE_INERTIA_TENSOR:
		BSDE_Mat3D2F_Copy(self->mass_imat, val);
		break;

	case BSDE_DIRECTION:
		V4D_COPY(self->dir, val);
		break;
	case BSDE_ROTATION_MATRIX:
		for(i=0; i<3; i++)for(j=0; j<3; j++)
			val[i*3+j]=self->xform[i*4+j];
		break;
	case BSDE_VERTICES:
		for(i=0; i<n; i++)val[i]=self->vecs[i];
		break;

	case BSDE_CONTACT_VELOCITY:
		V3D_COPY(self->u_wvel, val);
		break;
	case BSDE_ACCELERATION:
		V3D_COPY(self->u_acc, val);
		break;
	case BSDE_LOCAL_THRUST:
		V3D_COPY(self->u_racc, val);
		break;

	default:
		break;
	}
}

void BSDE_SetAttrFV(BSDE_Solid *self, int var, float *val, int n)
{
	switch(var)
	{
	case BSDE_ORIGIN:
		V3D_COPY(val, self->org);
		break;
	case BSDE_VELOCITY:
		V3D_COPY(val, self->vel);
		break;
	case BSDE_TORQUE:
		V3D_COPY(val, self->tvel);
		break;
	case BSDE_ANGULAR_VELOCITY:
		BSDE_WorldInertiaTensor(self);
		V3D_COPY(val, self->avel);
		BSDE_MatND_MatMult(self->wtmat, 3, 3, self->avel, 3, 1,
			self->tvel);
		break;
	case BSDE_ROTATION:
		V4D_COPY(val, self->rot);
		break;

	case BSDE_MINS:
		V3D_COPY(val, self->mins);
		break;
	case BSDE_MAXS:
		V3D_COPY(val, self->maxs);
		break;
	case BSDE_MASS_OFFSET:
		V3D_COPY(val, self->mass_offs);
		break;
	case BSDE_INERTIA_TENSOR:
		BSDE_Mat3F2D_Copy(val, self->mass_imat);
		BSDE_MatND_Inverse(self->mass_imat, self->mass_tmat, 3);
		break;

	case BSDE_DIRECTION:
		V4D_COPY(val, self->dir);
		break;

	case BSDE_CONTACT_VELOCITY:
		V3D_COPY(val, self->u_wvel);
		break;
	case BSDE_ACCELERATION:
		V3D_COPY(val, self->u_acc);
		break;
	case BSDE_LOCAL_THRUST:
		V3D_COPY(val, self->u_racc);
		break;

	default:
		break;
	}
}

void BSDE_GetAttrDV(BSDE_Solid *self, int var, double *val, int n)
{
	int i, j;

	switch(var)
	{
	case BSDE_ORIGIN:
		V3D_COPY(self->org, val);
		break;
	case BSDE_VELOCITY:
		V3D_COPY(self->vel, val);
		break;
	case BSDE_TORQUE:
		V3D_COPY(self->tvel, val);
		break;
	case BSDE_ANGULAR_VELOCITY:
		V3D_COPY(self->avel, val);
		break;
	case BSDE_ROTATION:
		V4D_COPY(self->rot, val);
		break;

	case BSDE_MINS:
		V3D_COPY(self->mins, val);
		break;
	case BSDE_MAXS:
		V3D_COPY(self->maxs, val);
		break;

	case BSDE_MASS_OFFSET:
		V3D_COPY(self->mass_offs, val);
		break;
	case BSDE_INERTIA_TENSOR:
		BSDE_Mat3D_Copy(self->mass_imat, val);
		break;

	case BSDE_DIRECTION:
		V4D_COPY(self->dir, val);
		break;
	case BSDE_ROTATION_MATRIX:
		for(i=0; i<3; i++)for(j=0; j<3; j++)
			val[i*3+j]=self->xform[i*4+j];
		break;
	case BSDE_VERTICES:
		for(i=0; i<n; i++)val[i]=self->vecs[i];
		break;

	case BSDE_CONTACT_VELOCITY:
		V3D_COPY(self->u_wvel, val);
		break;
	case BSDE_ACCELERATION:
		V3D_COPY(self->u_acc, val);
		break;
	case BSDE_LOCAL_THRUST:
		V3D_COPY(self->u_racc, val);
		break;

	default:
		break;
	}
}

void BSDE_SetAttrDV(BSDE_Solid *self, int var, double *val, int n)
{
	switch(var)
	{
	case BSDE_ORIGIN:
		V3D_COPY(val, self->org);
		break;
	case BSDE_VELOCITY:
		V3D_COPY(val, self->vel);
		break;
	case BSDE_TORQUE:
		V3D_COPY(val, self->tvel);
		break;
	case BSDE_ANGULAR_VELOCITY:
		BSDE_WorldInertiaTensor(self);
		V3D_COPY(val, self->avel);
		BSDE_MatND_MatMult(self->wtmat, 3, 3, self->avel, 3, 1,
			self->tvel);
		break;
	case BSDE_ROTATION:
		V4D_COPY(val, self->rot);
		break;

	case BSDE_MINS:
		V3D_COPY(val, self->mins);
		break;
	case BSDE_MAXS:
		V3D_COPY(val, self->maxs);
		break;
	case BSDE_MASS_OFFSET:
		V3D_COPY(val, self->mass_offs);
		break;
	case BSDE_INERTIA_TENSOR:
		BSDE_Mat3D_Copy(val, self->mass_imat);
		BSDE_MatND_Inverse(self->mass_imat, self->mass_tmat, 3);
		break;

	case BSDE_DIRECTION:
		V4D_COPY(val, self->dir);
		break;

	case BSDE_CONTACT_VELOCITY:
		V3D_COPY(val, self->u_wvel);
		break;
	case BSDE_ACCELERATION:
		V3D_COPY(val, self->u_acc);
		break;
	case BSDE_LOCAL_THRUST:
		V3D_COPY(val, self->u_racc);
		break;

	default:
		break;
	}
}

void BSDE_GetAttrIV(BSDE_Solid *self, int var, int *val, int n)
{
	int i;
	switch(var)
	{
	case BSDE_FACE_VERTICES:
		for(i=0; i<n; i++)val[i]=self->faces[i];
		break;
	case BSDE_FACE_COUNTS:
		for(i=0; i<n; i++)val[i]=self->facevecs[i];
		break;
	default:
		break;
	}
}

void BSDE_SetAttrIV(BSDE_Solid *self, int var, int *val, int n)
{
}

int BSDE_PointCollide(BSDE_Solid *self, double *point)
{
	double d;
	int i, j;

	i=0;
	switch(self->solidtype)
	{
	case BSDE_SOLID_AABB:
		i=1;
		for(j=0; j<3; j++)
		{
			d=point[j]-self->org[j];
			if((d<self->mins[j]) || (d>self->maxs[j]))i=0;
		}
		break;
	case BSDE_SOLID_OBB:
		i=BSDE_SolidOBB_ObbPointCollide(
			self->mins, self->maxs, self->xform, point);
		break;
	case BSDE_SOLID_SPHERE:
		i=1;
		d=V3F_DIST(self->org, point);
		if(d>self->radius)i=0;
		break;
	case BSDE_SOLID_CYLINDER:
		i=BSDE_SolidCylinder_PointCollide(
			self->height, self->radius, 0, self->xform,
			point);
		break;
	case BSDE_SOLID_CCYLINDER:
		i=BSDE_SolidCylinder_PointCollide(
			self->height, self->radius, 1, self->xform,
			point);
		break;
	case BSDE_SOLID_PLANE:
		i=BSDE_SolidPlane_PointCollide(
			self->dir, self->xform, point);
		break;
	case BSDE_SOLID_TRIMESH:
		i=BSDE_SolidOBB_ObbPointCollide(
			self->mins, self->maxs, self->xform, point);
		break;
	case BSDE_SOLID_HULL:
		i=BSDE_SolidHull_PointCollide(self, self->xform, point);
		break;
	default:
		break;
	}
	return(i);
}

void BSDE_PlaneExtents(BSDE_Solid *self, double *normal,
	double *min, double *max)
{
	double mins[3], maxs[3];
	double ext[2];
	double d, r;

	switch(self->solidtype)
	{
	case BSDE_SOLID_AABB:
		V3D_ADD(self->org, self->mins, mins);
		V3D_ADD(self->org, self->maxs, maxs);
		BSDE_SolidAABB_BoxPlaneExtents(mins, maxs, normal, min, max);
		break;
	case BSDE_SOLID_OBB:
		BSDE_SolidOBB_ObbPlaneExtents(self->mins, self->maxs,
			self->xform, normal, ext);
		*min=ext[0];
		*max=ext[1];
		break;
	case BSDE_SOLID_SPHERE:
		d=V3D_DOT(self->org, normal);
		*min=d-self->radius;
		*max=d+self->radius;
		break;
	case BSDE_SOLID_CYLINDER:
		r=BSDE_SolidCylinder_AxisRadius(
			self->height, self->radius, 0, self->xform, normal);
		d=V3D_DOT(self->org, normal);
		*min=d-r;
		*max=d+r;
		break;
	case BSDE_SOLID_CCYLINDER:
		r=BSDE_SolidCylinder_AxisRadius(
			self->height, self->radius, 1, self->xform, normal);
		d=V3D_DOT(self->org, normal);
		*min=d-r;
		*max=d+r;
		break;
	case BSDE_SOLID_PLANE:
		*min=-999999999;
		*max=999999999;
		break;

	case BSDE_SOLID_TRIMESH:
		BSDE_SolidBRep_ProjectLine(self->vecs, self->faces,
			self->n_vecs, self->n_faces, self->xform,
			normal, min, max);
		break;
	case BSDE_SOLID_HULL:
		BSDE_SolidHull_ProjectLine(self, self->xform,
			normal, min, max);
		break;
	case BSDE_SOLID_BSP:	//just be cheap...
		V3D_ADD(self->org, self->mins, mins);
		V3D_ADD(self->org, self->maxs, maxs);
		BSDE_SolidAABB_BoxPlaneExtents(mins, maxs, normal, min, max);
		break;

	default:
		break;
	}
}

void BSDE_PlaneExtentsPos(BSDE_Solid *self, double *pos, double *normal,
	double *min, double *max)
{
	double mins[3], maxs[3];
	double ext[2];
	double d, r;

	switch(self->solidtype)
	{
	case BSDE_SOLID_AABB:
		V3D_ADD(pos+3*4, self->mins, mins);
		V3D_ADD(pos+3*4, self->maxs, maxs);
		BSDE_SolidAABB_BoxPlaneExtents(mins, maxs, normal, min, max);
		break;
	case BSDE_SOLID_OBB:
		BSDE_SolidOBB_ObbPlaneExtents(self->mins, self->maxs,
			pos, normal, ext);
		*min=ext[0];
		*max=ext[1];
		break;
	case BSDE_SOLID_SPHERE:
		d=V3D_DOT(pos+3*4, normal);
		*min=d-self->radius;
		*max=d+self->radius;
		break;
	case BSDE_SOLID_CYLINDER:
//		r=BSDE_SolidCylinder_AxisRadius(
//			self->height, self->radius, 0, pos, normal);
		r=BSDE_SolidCylinder_AxisRadius(
			self->height, self->radius, 1, pos, normal);
		d=V3D_DOT(pos+3*4, normal);
		*min=d-r;
		*max=d+r;
		break;
	case BSDE_SOLID_CCYLINDER:
		r=BSDE_SolidCylinder_AxisRadius(
			self->height, self->radius, 1, pos, normal);
		d=V3D_DOT(pos+3*4, normal);
		*min=d-r;
		*max=d+r;
		break;
	case BSDE_SOLID_PLANE:
		*min=-999999999;
		*max=999999999;
		break;

	case BSDE_SOLID_TRIMESH:
		BSDE_SolidBRep_ProjectLine(self->vecs, self->faces,
			self->n_vecs, self->n_faces, pos,
			normal, min, max);
		break;
	case BSDE_SOLID_HULL:
		BSDE_SolidHull_ProjectLine(self, pos, normal, min, max);
		break;
	case BSDE_SOLID_BSP:	//just be cheap...
		V3D_ADD(pos+3*4, self->mins, mins);
		V3D_ADD(pos+3*4, self->maxs, maxs);
		BSDE_SolidAABB_BoxPlaneExtents(mins, maxs, normal, min, max);
		break;

	default:
		break;
	}
}

void BSDE_PlaneMoveExtents(BSDE_Solid *self, double *normal,
	double *min, double *max)
{
	double m, n, o, p;
	BSDE_PlaneExtentsPos(self, self->xform, normal, &m, &n);
	BSDE_PlaneExtentsPos(self, self->e_xform, normal, &o, &p);

	*min=(m<o)?m:o;
	*max=(n>p)?n:p;
}

void BSDE_PlaneFastMoveExtents(BSDE_Solid *self, double *normal,
	double *min, double *max)
{
	BSDE_SolidAABB_BoxPlaneExtents(self->mmins, self->mmaxs,
		normal, min, max);
}
