#ifndef BSDE_SOLID_H
#define BSDE_SOLID_H

typedef struct BSDE_Solid_s BSDE_Solid;
typedef struct BSDE_Contact_s BSDE_Contact;
typedef struct BSDE_Joint_s BSDE_Joint;
typedef struct BSDE_World_s BSDE_World;
typedef struct BSDE_QuadNode_s BSDE_QuadNode;
typedef struct BSDE_BinaryNode_s BSDE_BinaryNode;

struct BSDE_Solid_s {
//solid management
BSDE_Solid *qnext;	//next in quad/oct tree
BSDE_Solid *chain;	//chained solid
BSDE_BinaryNode *node;	//linked in node

BSDE_World *world;	//world solid belongs to
int handle;		//handle allocated to this solid

BSDE_Joint *joint;	//first joint
void *data;		//solid specific data

//solid

int solidtype;	//solid type
int moveflags;	//flags effecting the movement of an object
int stateflags;	//internally managed flags

//general data
double mins[3], maxs[3];	//AABB or OBB extents
double height, radius;	//height and radius (depend on type)
double dir[4];		//direction (planes)

//general static properties
double mass;
double volume;		//internal volume
double mass_offs[3];	//XYZ, mass offset
double mass_imat[9];	//inertial matrix
double mass_tmat[9];	//inverse inertial matrix
double elastic;		//surface elastic
double friction;	//surface friction
double viscosity;	//fluid viscosity

//temp state
double mmins[3];	//move min extents
double mmaxs[3];	//move max extents
double ivel[3];		//impulse velocity
double itvel[3];	//impulse torque
double avel[3];		//angular velocity
double wtmat[9];	//world inverse inertial tensor, I^-1
double wimat[9];	//world inertial tensor, I

//general state
double org[4];		//XYZ0
double rot[4];		//quaternion
double vel[4];		//XYZ0
double tvel[4];		//XYZ0, torque/inertia
double xform[16];	//current transform

//predicted ending state
double e_org[4];		//XYZ0
double e_rot[4];		//quaternion
double e_vel[4];		//XYZ0
double e_tvel[4];	//XYZ0, torque/inertia
double e_xform[16];	//predicted transform

//special state (user forces, ...)
double u_wvel[3];	//desired contact velocity
double u_acc[3];	//acceleration
double u_racc[3];	//localspace acceleration (thrust)
double u_resv[14];	//space for more vectors
double u_wpow;		//contact velocity power

//solidtype related and similar
int n_vecs;
int n_faces;
double *vecs;	//vertices
double *norms;	//face normals (x y z d)
int *faces;	//faces/triangle vertex indices
int *facevecs;	//vertices/face (hulls)
double *evecs;	//edge normals

double *e_vecs;		//predicted vertices (trimesh)
double *e_evecs;	//predicted edge vectors (trimesh)
};

struct BSDE_Contact_s {
BSDE_Solid *obja;	//first colliding object
BSDE_Solid *objb;	//second colliding object

double org[3];		//intersection point
double norm[3];		//collision normal
double depth;		//penetration depth
//norm and depth define the amount obja should move
//norm and -depth define the amount objb should move
};

struct BSDE_Joint_s {
BSDE_Solid *obja;	//first object
BSDE_Solid *objb;	//second object
BSDE_Joint *linka;	//next joint for obja
BSDE_Joint *linkb;	//next joint for objb

int type;
int flags;

double orga[3];		//joint position 0 (a's local space)
double orgb[3];		//joint position 0 (b's local space)
double dira[3];		//joint direction 0 (a's local space)
double dirb[3];		//joint direction 0 (b's local space)
double dir1a[3];	//joint direction 1 (a's local space)
double dir1b[3];	//joint direction 1 (b's local space)

double org1a[3];	//joint position 1 (a's local space)
double org1b[3];	//joint position 1 (b's local space)
double org2a[3];	//joint position 2 (a's local space)
double org2b[3];	//joint position 2 (b's local space)
double org3a[3];	//joint position 3 (a's local space)
double org3b[3];	//joint position 3 (b's local space)
};


struct BSDE_World_s {
BSDE_Solid **solids;
int n_solids;

BSDE_Joint **joints;
int n_joints;

BSDE_Solid **stack;
int stackpos, stacksize;

BSDE_Contact **contact;
//BSDE_QuadNode *quad;
void *root;		//root of partitioning tree

double gravity[3];	//gravity vector
double maxstep;		//maximum step size
double minstep;		//minimum step size
double maxframe;	//maximum frame size

double worldtime;	//accumulated world time
double steptime;	//remaining step time
double realtime;	//approximate "realtime"

int flags;		//world-related flags
int passes;		//passes per frame

int *tr_arr;	//trigger array [self, other]
int tr_num;	//size of trigger array
int tr_cur;	//current spot in trigger array

void *t_arr_hs;	//self handles
void *t_arr_ho;	//other handles
void *t_arr_o;	//(x, y, z) origins
void *t_arr_n;	//(x, y, z, d) normals
int t_str_hs;	//self handle stride
int t_str_ho;	//other handle stride
int t_str_o;	//origin stride
int t_str_n;	//normal stride
int t_ty_hs;	//self handle type
int t_ty_ho;	//other handle type
int t_ty_o;	//origin type
int t_ty_n;	//normal type

int t_num;	//max size of contact array
int t_cur;	//current index in contact array
};

struct BSDE_QuadNode_s {
BSDE_QuadNode *node[4];
double center[4];
BSDE_Solid *first;	//first solid in node
};

struct BSDE_BinaryNode_s {
BSDE_BinaryNode *node[2], *up;
double norm[4];
BSDE_Solid *first;	//first solid in node
};

#endif
