/*
Convex Hulls

Each will contain:
A number of face normals defining the solid;
The number of points on each face;
A list of points for each face;
A list of points for the hull.

Normals: N=(x y z w)=(Nxyz w)
Where: w=O*Nxyz
 */

#include "bsde_internal.h"


void BSDE_SolidHull_MakePlaneFace(double *norm, double *pts)
{
	double v0[3], v1[3], v2[3];

//	v0[0]=norm[1]; v0[1]=norm[2]; v0[2]=norm[0];
//	v1[0]=norm[2]; v1[1]=norm[0]; v1[2]=norm[1];

	BSDE_PlaneD_AdjacentNormals(norm, v0, v1);

//	printf("  [(%g %g %g) (%g %g %g) (%g %g %g)]",
//		v0[0], v0[1], v0[2],
//		v1[0], v1[1], v1[2],
//		norm[0], norm[1], norm[2]);

//	V3D_SCALEADDSCALE(v0, -999, v1, -999, pts+(0*3));
//	V3D_SCALEADDSCALE(v0,  999, v1, -999, pts+(1*3));
//	V3D_SCALEADDSCALE(v0,  999, v1,  999, pts+(2*3));
//	V3D_SCALEADDSCALE(v0, -999, v1,  999, pts+(3*3));
//	V3D_ADDSCALE(pts+(1*3), norm, norm[3], pts+(1*3));
//	V3D_ADDSCALE(pts+(2*3), norm, norm[3], pts+(2*3));
//	V3D_ADDSCALE(pts+(3*3), norm, norm[3], pts+(3*3));

	V3D_SCALEADDSCALE(v0, -999, v1, -999, v2);
	V3D_ADDSCALE(v2, norm, norm[3], pts+(0*3));
	V3D_SCALEADDSCALE(v0,  999, v1, -999, v2);
	V3D_ADDSCALE(v2, norm, norm[3], pts+(1*3));
	V3D_SCALEADDSCALE(v0,  999, v1,  999, v2);
	V3D_ADDSCALE(v2, norm, norm[3], pts+(2*3));
	V3D_SCALEADDSCALE(v0, -999, v1,  999, v2);
	V3D_ADDSCALE(v2, norm, norm[3], pts+(3*3));
}

int BSDE_SolidHull_ClipFace(double *norm,
	double *ipts, double *opts, int num)
{
	int i, j, k, l;

	//first outside
	for(i=0; i<num; i++)if(V3D_NDOT(ipts+(i*3), norm)>0)break;

	if(i==num)	//nothing to clip
	{
		for(i=0; i<num*3; i++)opts[i]=ipts[i];
		return(num);	//nothing to clip
	}

	//first inside
	j=i;
	while(1)
	{
		if(V3D_NDOT(ipts+(j*3), norm)<=0)break;
		j=(j+1)%num;
		if(j==i)return(0);	//everything clipped
	}

	//copy inside
	i=j;
	l=0;
	while(1)
	{
		V3D_COPY(ipts+(j*3), opts+(l*3));
		l++;

		k=(j+1)%num;
		if(V3D_NDOT(ipts+(k*3), norm)>0)break;
		j=k;
	}

	//exit point
	if(V3D_NDOT(ipts+(j*3), norm)<0)
	{
		BSDE_PlaneD_LinePlaneIntersect(
			ipts+(j*3), ipts+(k*3), norm, opts+(l*3));
		l++;
	}

	j=k;
	while(1)
	{
		k=(j+1)%num;
		if(V3D_NDOT(ipts+(k*3), norm)<=0)break;
		j=k;
	}

	//entry point
	if(V3D_NDOT(ipts+(k*3), norm)<0)
	{
		BSDE_PlaneD_LinePlaneIntersect(
			ipts+(j*3), ipts+(k*3), norm, opts+(l*3));
		l++;
	}

	return(l);
}

void BSDE_SolidHull_MakeHullFaces(BSDE_Solid *obj)
{
	static double vecbuf[64*64*3];
	static double pointbuf[64*3], pointbuf2[64*3];
	static int facebuf[64*64];
	static int facevbuf[64];
	int i, j, k, l, n, nf, nv;
	int p0, p1, p2;
	double *v0, *v1, *v2;
	double dv0[3], dv1[3], dv2[3];
	double f, g, h;

	n=0;
	nv=0;
	nf=0;

//	printf("build");
	for(i=0; i<obj->n_faces; i++)
	{
		BSDE_SolidHull_MakePlaneFace(obj->norms+(i*4), pointbuf);
		l=4;

		for(j=0; j<obj->n_faces; j++)
		{
			if(i==j)continue;

			l=BSDE_SolidHull_ClipFace(obj->norms+(j*4),
				pointbuf, pointbuf2, l);
			for(k=0; k<(l*3); k++)pointbuf[k]=pointbuf2[k];
		}
		if(!l)continue;

//		printf("%d:%d", i, l);

		for(j=0; j<l; j++)
		{
			for(k=0; k<nv; k++)
				if(V3D_DIST(pointbuf+j*3,
					vecbuf+k*3)<0.00001)
						break;
			if(k==nv)
//			if(1)
			{
				V3D_COPY(pointbuf+j*3, vecbuf+k*3);
				nv++;
			}

			facebuf[n++]=k;
		}

		facevbuf[nf]=l;
		nf++;
	}
//	printf("\n");

//	printf("facevecs");
//	for(i=0; i<nf; i++)printf(" %d:%d", i, facevbuf[i]);
//	printf("\n");

	if(nf<obj->n_faces)
	{
		printf("MakeHullFaces: %d faces clipped away\n",
			obj->faces-nf);
	}

//	obj->n_faces=nf;

	obj->n_vecs=nv;
	obj->vecs=malloc(nv*3*sizeof(double));
	memcpy(obj->vecs, vecbuf, nv*3*sizeof(double));

	obj->faces=malloc(n*sizeof(int));
	memcpy(obj->faces, facebuf, n*sizeof(int));

	obj->facevecs=malloc(nf*sizeof(int));
	memcpy(obj->facevecs, facevbuf, nf*sizeof(int));

	k=0;
	h=0;
	for(i=0; i<nf; i++)
	{
		l=facevbuf[i];
		for(j=1; j<(l-1); j++)
		{
			p0=facebuf[k];
			p1=facebuf[k+j];
			p2=facebuf[k+j+1];
			if((p0==p1) || (p1==p2) || (p2==p0))continue;

			v0=vecbuf+p0*3;
			v1=vecbuf+p1*3;
			v2=vecbuf+p2*3;

			V3D_SUB(v1, v0, dv0);
			V3D_SUB(v2, v0, dv1);

			V3D_CROSS(dv0, dv1, dv2);
			f=V3D_NORMALIZE(dv2, dv2);
			g=f*0.5;

			V3D_CROSS(dv0, v0, dv2);
			f=V3D_NORMALIZE(dv2, dv2);
			g*=f*0.5;

			V3D_CROSS(dv1, v0, dv2);
			f=V3D_NORMALIZE(dv2, dv2);
			g*=f*0.5;

			h+=g;
		}
		k+=l;
	}

	obj->volume=h;
}

void BSDE_SolidHull_ComputeVolume(BSDE_Solid *obj)
{
	int i, j;
}

void BSDE_SolidHull_ProjectLine(BSDE_Solid *obj,
	double *xform, double *norm, double *min, double *max)
{
	double pt[3];
	double m, n, f;
	int i;

	m=999999999.0;
	n=-999999999.0;
	
	for(i=0; i<obj->n_vecs; i++)
	{
		BSDE_PlaneD_TransformPoint(obj->vecs+i*3, xform, pt);
		f=V3D_DOT(pt, norm);
		if(f<m)m=f;
		if(f>n)n=f;
	}

	*min=m;
	*max=n;
}

void BSDE_SolidHull_Box(BSDE_Solid *obj,
	double *xform, double *mins, double *maxs)
{
	double pt[3];
	int i, j;

	V3D_CONST(mins, BSDE_M_BIGVAL);
	V3D_CONST(maxs, -BSDE_M_BIGVAL);
	
	for(i=0; i<obj->n_vecs; i++)
	{
		BSDE_PlaneD_TransformPoint(obj->vecs+i*3, xform, pt);
		for(j=0; j<3; j++)
		{
			mins[j]=(mins[j]<pt[j])?mins[j]:pt[j];
			maxs[j]=(maxs[j]>pt[j])?maxs[j]:pt[j];
		}
	}
}

void BSDE_SolidHull_BoxMove(BSDE_Solid *obj,
	double *s_xform, double *e_xform, double *mins, double *maxs)
{
	double m0[3], n0[3], m1[3], n1[3];

	BSDE_SolidHull_Box(obj, s_xform, m0, n0);
	BSDE_SolidHull_Box(obj, e_xform, m1, n1);
	mins[0]=(m0[0]<m1[0])?m0[0]:m1[0];
	maxs[0]=(n0[0]>n1[0])?n0[0]:n1[0];
	mins[1]=(m0[1]<m1[1])?m0[1]:m1[1];
	maxs[1]=(n0[1]>n1[1])?n0[1]:n1[1];
	mins[2]=(m0[2]<m1[2])?m0[2]:m1[2];
	maxs[2]=(n0[2]>n1[2])?n0[2]:n1[2];
}

int BSDE_SolidHull_PointCollide(BSDE_Solid *obj, double *xform,
	double *org)
{
	double norm[4];
	double f;
	int i;

	for(i=0; i<obj->n_faces; i++)
	{
		BSDE_PlaneD_TransformNormal(obj->norms+i*4, xform, norm);
		f=V3D_NDOT(org, norm);
		if(f>0)return(0);
	}
	return(1);
}

void BSDE_SolidHull_NearestPoint(BSDE_Solid *obj, double *xform,
	double *org, double *pt)
{
	double norm[4];
	double f;
	int i;

	V3D_COPY(org, pt);
	for(i=0; i<obj->n_faces; i++)
	{
		BSDE_PlaneD_TransformNormal(obj->norms+i*4, xform, norm);
		f=V3D_NDOT(pt, norm);
		if(f>0)
		{
			V3D_ADDSCALE(pt, norm, -f, pt);
		}
	}
}

int BSDE_SolidHull_CheckCollide(BSDE_Solid *aobj, BSDE_Solid *bobj,
	double *axform, double *bxform)
{
	double norm[4];
	double m, n;
	int i;

	for(i=0; i<aobj->n_faces; i++)
	{
		BSDE_PlaneD_TransformNormal(aobj->norms+i*4, axform, norm);
		BSDE_SolidHull_ProjectLine(bobj, bxform, norm, &m, &n);
		if(m>=norm[3])return(0);
	}

	for(i=0; i<bobj->n_faces; i++)
	{
		BSDE_PlaneD_TransformNormal(bobj->norms+i*4, bxform, norm);
		BSDE_SolidHull_ProjectLine(aobj, axform, norm, &m, &n);
		if(m>=norm[3])return(0);
	}

	return(1);
}

int BSDE_SolidHull_CheckObbCollide(BSDE_Solid *aobj, BSDE_Solid *bobj,
	double *axform, double *bxform)
{
	double norm[4], ext[4];
	double m, n;
	int i;

	for(i=0; i<aobj->n_faces; i++)
	{
		BSDE_PlaneD_TransformNormal(aobj->norms+i*4, axform, norm);
		BSDE_SolidOBB_ObbPlaneExtents(bobj->mins, bobj->maxs,
			bxform, norm, ext);
		if(ext[0]>=norm[3])return(0);
	}

	for(i=0; i<3; i++)
	{
		V3D_COPY(bxform+i*4, norm);
		BSDE_SolidHull_ProjectLine(aobj, axform, norm, &m, &n);
		BSDE_SolidOBB_ObbPlaneExtents(bobj->mins, bobj->maxs,
			bxform, norm, ext);
		if((ext[0]>=n) || (m>=ext[1]))
			return(0);
	}

	return(1);
}

int BSDE_SolidHull_CheckSphereCollide(BSDE_Solid *aobj, BSDE_Solid *bobj,
	double *axform, double *bxform, double *aorg, double *borg)
{
	double norm[4], ext[4];
	double pt[3];
	double m, n;
	int i;

	BSDE_SolidHull_NearestPoint(aobj, axform, borg, pt);
	if(V3D_DIST(borg, pt)>=bobj->radius)
		return(0);

#if 0
	for(i=0; i<aobj->n_faces; i++)
	{
		BSDE_PlaneD_TransformNormal(aobj->norms+i*4, axform, norm);
		m=V3D_NDOT(borg, norm)-bobj->radius;
		if(m>=norm[3])return(0);
	}
#endif

	return(1);
}

int BSDE_SolidHull_CheckCylinderCollide(BSDE_Solid *aobj, BSDE_Solid *bobj,
	double *axform, double *bxform, double *aorg, double *borg)
{
	double f;
	double pt[3], pt2[3];
	double m, n;
	int i;

	f=V3D_DOT(aorg, bxform+2*4)-V3D_DOT(borg, bxform+2*4);

	if(f>(bobj->height*0.5))f=(bobj->height*0.5);
	if(f<(-(bobj->height*0.5)))f=-(bobj->height*0.5);

	pt[0]=borg[0]+f*bxform[2*4+0];
	pt[1]=borg[1]+f*bxform[2*4+1];
	pt[2]=borg[2]+f*bxform[2*4+2];

	BSDE_SolidHull_NearestPoint(aobj, axform, pt, pt2);
	if(V3D_DIST(pt, pt2)>=bobj->radius)
		return(0);

	if(bobj->solidtype==BSDE_SOLID_CYLINDER)
	{
		f=V3D_DOT(pt2, bxform+2*4)-V3D_DOT(borg, bxform+2*4);
		if(f>(bobj->height*0.5))return(0);
		if(f<(-(bobj->height*0.5)))return(0);
	}

	return(1);
}

int BSDE_SolidHull_CheckTriMeshCollide(BSDE_Solid *aobj, BSDE_Solid *bobj,
	double *axform, double *bxform)
{
	static double norms[64*4];
	double *v0, *v1, *v2;
	double norm[4], v[3];
	double m, n;
	int i, j;

//	printf("mesh/hull\n");

	BSDE_SolidBRep_Predict(bobj);

	for(i=0; i<aobj->n_faces; i++)
		BSDE_PlaneD_TransformNormal(
			aobj->norms+i*4, axform, norms+i*4);

	//try to rule out by seperating axis
	for(i=0; i<aobj->n_faces; i++)
	{
		BSDE_SolidBRep_ProjectLine3(bobj, norms+i*4, &m, &n);
		if(m>=norms[i*4+3])return(0);
	}

//	printf("mesh/hull 1\n");

	return(1);

	//verify that at least 1 face is inside the hull
	for(i=0; i<bobj->n_faces; i++)
	{
		v0=bobj->e_vecs+bobj->faces[i*3+0]*3;
		v1=bobj->e_vecs+bobj->faces[i*3+0]*3;
		v2=bobj->e_vecs+bobj->faces[i*3+0]*3;

		for(j=0; j<aobj->n_faces; j++)
		{
			v[0]=V3D_NDOT(v0, norms+i*4);
			v[1]=V3D_NDOT(v1, norms+i*4);
			v[2]=V3D_NDOT(v2, norms+i*4);

			if((v[0]>=0) && (v[1]>=0) && (v[2]>=0))
				break;
		}

		if(j==aobj->n_faces)
			return(1);
	}

	return(0);
}

int BSDE_SolidHull_CollidePoints(BSDE_Solid *obj,
	double *ipts, double *opts, int npts)
{
	double f;
	int i, j, k, l;

	l=0;
	for(i=0; i<npts; i++)
	{
		k=0;
		for(j=0; j<obj->n_faces; j++)
		{
			f=V3D_NDOT(ipts+i*3, obj->norms+j*4);
			if(f>0)
			{
				k=1;
				break;
			}
		}
		if(k)continue;

		V3D_COPY(ipts+i*3, opts+l*3);
		l++;
	}

	return(l);
}

void BSDE_SolidHull_CollideOrigin(BSDE_Solid *aobj, BSDE_Solid *bobj,
	double *org)
{
	double vecs[256*3];
	int i, j;

	i=BSDE_SolidHull_CollidePoints(aobj,
		bobj->vecs, vecs, bobj->n_vecs);
	j=BSDE_SolidHull_CollidePoints(bobj,
		aobj->vecs, vecs+i*3, aobj->n_vecs);

	i=i+j;
	V3D_ZERO(org);
	for(j=0; j<i; j++)
	{
		V3D_ADD(org, vecs+j*3, org);
	}
	V3D_SCALE(org, 1.0/i, org);
}


int BSDE_SolidHull_ContactHull(BSDE_Solid *aobj, BSDE_Solid *bobj,
	double *axform, double *bxform, double *aorg, double *borg,
	double *org, double *norm, double *dist)
{
	static double anv[64*4], bnv[64*4];

	double bn[4], bo[3], tn[4], pt[3];
	double bd;
	double m, n, o, p, d, f;
	int i, j, k;

#if 0
	for(i=0; i<aobj->n_faces; i++)
		BSDE_PlaneD_TransformNormal(aobj->norms+i*4, axform, anv+i*4);
	for(i=0; i<bobj->n_faces; i++)
		BSDE_PlaneD_TransformNormal(bobj->norms+i*4, bxform, bnv+i*4);

	bd=999999;
	for(i=0; i<aobj->n_faces; i++)
	{
		BSDE_PlaneExtents(bobj, anv+i*4, &m, &n);
		if((m>anv[i*4+3]) || (n<anv[i*4+3]))continue;

		d=V3D_DOT(bobj->vel, anv+i*4);
		if(d<bd)
		{
			V4D_COPY(anv+i*4, bn);
			bd=d;
		}
	}
	for(i=0; i<bobj->n_faces; i++)
	{
		BSDE_PlaneExtents(aobj, bnv+i*4, &m, &n);
		if((m>bnv[i*4+3]) || (n<bnv[i*4+3]))continue;

		d=V3D_DOT(aobj->vel, bnv+i*4);
		if(d<bd)
		{
			V4D_COPY(bnv+i*4, bn);
			bd=d;
		}
	}

#if 1
	if(V3D_DOT(aorg, bn)<V3D_DOT(borg, bn))
	{
		V4D_SCALE(bn, -1, bn);
	}
	V3D_COPY(bn, norm);
#endif

#if 1
	V3D_ZERO(bo);
	k=0;

	for(i=0; i<aobj->n_vecs; i++)
	{
		BSDE_PlaneD_TransformPoint(aobj->vecs+i*3, axform, pt);

		for(j=0; j<bobj->n_faces; j++)
			if(V3D_NDOT(pt, bnv+j*4)>0)break;
		if(j==bobj->n_faces)
		{
			V3D_ADD(bo, pt, bo);
			k++;
		}
	}

	for(i=0; i<bobj->n_vecs; i++)
	{
		BSDE_PlaneD_TransformPoint(bobj->vecs+i*3, bxform, pt);

		for(j=0; j<aobj->n_faces; j++)
			if(V3D_NDOT(pt, anv+j*4)>0)break;
		if(j==aobj->n_faces)
		{
			V3D_ADD(bo, pt, bo);
			k++;
		}
	}

	if(k)
	{
		V3D_SCALE(bo, 1.0/k, org);
	}else
	{
		d=V3D_DOT(borg, norm);
		V3D_ADDSCALE(borg, norm, ((m+n)/2)-d, org);
	}
#endif

	BSDE_PlaneExtents(aobj, norm, &m, &n);
	BSDE_PlaneExtents(bobj, norm, &o, &p);
	m=(m>o)?m:o; n=(n<p)?n:p;

//	d=V3D_DOT(borg, norm);
//	V3D_ADDSCALE(borg, norm, ((m+n)/2)-d, org);
	*dist=n-m;
	return(1);
#endif


	V3D_SUB(aorg, borg, bn);
	V3D_NORMALIZE(bn, norm);

	BSDE_PlaneExtents(aobj, norm, &m, &n);
	BSDE_PlaneExtents(bobj, norm, &o, &p);
	m=(m>o)?m:o; n=(n<p)?n:p;

	d=V3D_DOT(borg, norm);
	V3D_ADDSCALE(borg, norm, ((m+n)/2)-d, org);
	*dist=n-m;
	return(1);
}

int BSDE_SolidHull_ContactGeneric(BSDE_Solid *aobj, BSDE_Solid *bobj,
	double *axform, double *bxform, double *aorg, double *borg,
	double *org, double *dir, double *dist)
{
	double norm[4], bn[4], bo[3];
	double bd;
	double m, n, o, p, d, f;
	int i;

	if(bobj->solidtype==BSDE_SOLID_SPHERE)
	{
		BSDE_SolidHull_NearestPoint(aobj, axform, borg, bo);
		if(V3D_DIST(borg, bo)<0.001)V3D_COPY(aorg, bo);

//		V3D_SUB(borg, bo, dir);
		V3D_SUB(bo, borg, dir);
		V3D_NORMALIZE(dir, dir);

		V3D_SUB(bo, borg, bn);
		d=bobj->radius/V3D_LEN(bn);
		V3D_SCALE(bn, d, bn);
		V3D_ADD(borg, bn, org);

		BSDE_PlaneExtents(aobj, dir, &m, &n);
		*dist=V3D_DOT(org, dir)-m;

		return(1);
	}

	V3D_SUB(aorg, borg, bn);
	V3D_NORMALIZE(bn, dir);

	BSDE_PlaneExtents(aobj, dir, &m, &n);
	BSDE_PlaneExtents(bobj, dir, &o, &p);
	m=(m>o)?m:o; n=(n<p)?n:p;

	d=V3D_DOT(borg, dir);
	V3D_ADDSCALE(borg, dir, ((m+n)/2)-d, org);
	*dist=n-m;
	return(1);

#if 0
	V3D_ZERO(bn);
	V3D_ZERO(bo);
	bd=999999;

	for(i=0; i<aobj->n_faces; i++)
	{
		BSDE_PlaneD_TransformNormal(aobj->norms+i*4, axform, norm);
		BSDE_PlaneExtents(bobj, norm, &m, &n);
		d=norm[3]-m;
		if(d<0)return(0);

		if(d<bd)
		{
			V4D_COPY(norm, bn);
			bd=d;
		}
	}

	d=V3D_NDOT(borg, bn);
	V3D_ADDSCALE(borg, bn, -d, org);
	V3D_SCALE(bn, -1, dir);
	*dist=bd;

	return(1);
#endif

#if 0
	V3D_COPY(borg, org);
	V3D_ZERO(bn);

	for(i=0; i<aobj->n_faces; i++)
	{
		BSDE_PlaneD_TransformNormal(aobj->norms+i*4, axform, norm);
		BSDE_PlaneExtents(bobj, norm, &m, &n);
		if((m<norm[3]) && (n>norm[3]))
		{
			d=norm[3]-m;
			f=V3D_DOT(org, norm)-m+d*0.5;
			V3D_ADDSCALE(org, norm, -f, org);

			V3D_ADD(bn, norm, bn);
		}
	}

	V3D_NORMALIZE(bn, bn);
	V3D_SCALE(bn, -1, bn);

	BSDE_PlaneExtents(bobj, norm, &m, &n);
	BSDE_SolidHull_ProjectLine(aobj, axform, bn, &o, &p);
	m=(m>o)?m:o;
	n=(n<p)?n:p;

	*dist=n-m;
#endif
}

void BSDE_SolidHull_HullPlaneIntersect(BSDE_Solid *obj, double *pos,
	double *norm, double *io)
{
	double bo[3], pt[3];
	double d;
	int i, j;

	V3D_ZERO(bo);
	j=0;

	for(i=0; i<obj->n_vecs; i++)
	{
		BSDE_PlaneD_TransformPoint(obj->vecs+i*3, pos, pt);
		d=V3D_NDOT(pt, norm);
		if(d<0)
		{
			V3D_ADD(bo, pt, bo);
			j++;
		}
	}

	if(!j)j++;
	V3D_SCALE(bo, 1.0/j, io);
}
